# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
"""Utilities used during AutoML training."""
import ast
import logging
from typing import Any, Callable, Dict, List, Optional, Tuple, Union, cast

import azureml.dataprep as dprep
import numpy as np
import pandas as pd
from azureml._common._error_definition import AzureMLError
from azureml._common._error_definition.user_error import ArgumentBlankOrEmpty, ArgumentOutOfRange
from azureml._tracing._tracer_factory import get_tracer
from azureml.automl.runtime import (_data_transformation_utilities,
                                    dataprep_utilities,
                                    frequency_fixer,
                                    short_grain_padding)
from azureml.automl.runtime._automl_settings_utilities import rule_based_validation
from azureml.automl.runtime._data_definition.raw_experiment_data import RawExperimentData
from azureml.automl.runtime._ml_engine import ml_engine
from azureml.automl.runtime._ml_engine.validation import common_data_validations
from azureml.automl.runtime.data_context import TransformedDataContext
from azureml.automl.runtime.featurizer.transformer.timeseries import forecasting_heuristic_utils
from azureml.automl.runtime.frequency_fixer import fix_data_set_regularity_may_be
from azureml.automl.runtime.shared import utilities as runtime_utilities
from azureml.automl.runtime.shared._cv_splits import _CVSplits
from azureml.automl.runtime.shared.cache_store import CacheStore
from azureml.automl.runtime.shared.datasets import ClientDatasets, DatasetBase, SubsampleCacheStrategy
from azureml.automl.runtime.shared.streaming_dataset import DatasetMetadataKeys, StreamingDataset
from azureml.automl.runtime.shared.types import CoreDataInputType, DataInputType, DataSingleColumnInputType
from azureml.automl.runtime.shared.utilities import _get_num_unique
from azureml.automl.runtime.streaming_data_context import StreamingTransformedDataContext
from azureml.dataprep import DataPrepException as DprepException
from azureml.dataprep.api.dataflow import DataflowValidationError
from scipy import sparse
from sklearn.base import TransformerMixin

from azureml.automl.core.automl_base_settings import AutoMLBaseSettings
from azureml.automl.core.shared import constants, logging_utilities, utilities
from azureml.automl.core.shared._diagnostics.automl_error_definitions import (ConflictingValueForArguments,
                                                                              DataShapeMismatch,
                                                                              InconsistentColumnTypeInTrainValid,
                                                                              IndistinctLabelColumn,
                                                                              InvalidArgumentWithSupportedValues,
                                                                              InvalidInputDatatype,
                                                                              InvalidValuesInCVSplitColumn,
                                                                              MissingColumnsInData,
                                                                              OverlappingYminYmax)
from azureml.automl.core.shared._diagnostics.contract import Contract
from azureml.automl.core.shared._diagnostics.validation import Validation
from azureml.automl.core.shared.constants import NumericalDtype, TelemetryConstants
from azureml.automl.core.shared.exceptions import ConfigException, DataException, UserException, ValidationException
from azureml.automl.core.shared.reference_codes import ReferenceCodes

logger = logging.getLogger(__name__)
tracer = get_tracer(__name__)


class LargeDatasetLimit:
    """Constants for limiting large datasets."""

    MAX_ROWS_TO_SUBSAMPLE = 100000

    # Number of rows to  use for doing validations on the data
    VALIDATION_SUBSAMPLE_SIZE = 5000


MASKED = '[Masked]'


def auto_block_models(input_data, automl_settings):
    """
    Add appropriate files to blocked_models automatically.

    :param input_data:
    :param automl_settings: The settings used for this current run.
    :return:
    """
    # Only enable auto-block if user didn't add any models to the allowed list
    if automl_settings.auto_blacklist and not automl_settings.whitelist_models:
        X = input_data['X']
        if sparse.issparse(X) or X.shape[0] > constants.MAX_SAMPLES_AUTOBLOCK:
            if automl_settings.blacklist_algos is None:
                automl_settings.blacklist_algos = \
                    constants.MAX_SAMPLES_AUTOBLOCKED_ALGOS
            else:
                for blacklist_algo in constants.MAX_SAMPLES_AUTOBLOCKED_ALGOS:
                    if blacklist_algo not in automl_settings.blacklist_algos:
                        automl_settings.blacklist_algos.append(blacklist_algo)
            automl_settings.blacklist_samples_reached = True
            automl_settings._validate_model_filter_lists()


def set_task_parameters(y: DataSingleColumnInputType, automl_settings: AutoMLBaseSettings) -> None:
    """
    Set this task's parameters based on some heuristics if they aren't provided.

    TODO: Move this code into AutoML settings or something. Client shouldn't have to think about this stuff.

    :param y: The list of possible output values.
    :param automl_settings: The settings used for this current run.
    :return: None.
    :raises: `ValidationException` in case of type errors. `DataException` in case of single valued target column
    """
    if automl_settings.enable_streaming:
        # Do not apply these heuristics if streaming is enabled.
        return

    Contract.assert_true(automl_settings.task_type in {constants.Tasks.CLASSIFICATION, constants.Tasks.REGRESSION},
                         "Invalid task type ({}) specified.".format(automl_settings.task_type))

    if automl_settings.task_type == constants.Tasks.CLASSIFICATION:
        #  Guess number of classes if the user did not explicitly provide it
        if not automl_settings.num_classes or not isinstance(
                automl_settings.num_classes, int):
            automl_settings.num_classes = _get_num_unique(pd.Series(y), ignore_na=True)
        return

    if automl_settings.task_type == constants.Tasks.REGRESSION:
        y_datatype = runtime_utilities._get_column_data_type_as_str(y)
        if not utilities._check_if_column_data_type_is_numerical(y_datatype):
            raise ValidationException._with_error(AzureMLError.create(
                InvalidInputDatatype,
                input_type=y_datatype,
                reference_code=ReferenceCodes._VALIDATE_TARGET_COLUMN_INCORRECT_TYPE,
                supported_types=", ".join(list(NumericalDtype.FULL_SET)),
                target='y'
            ))

        #  Guess min and max of y if the user did not explicitly provide it
        y_min = automl_settings.y_min
        y_max = automl_settings.y_max
        if not automl_settings.y_min:
            y_min = np.nanmin(y)
        if not automl_settings.y_max:
            y_max = np.nanmax(y)
        if y_max is not None and y_min is not None and y_max == y_min:
            raise DataException._with_error(AzureMLError.create(
                OverlappingYminYmax, target="y/y_min/y_max", value=str(y_min))
            )


def format_training_data(
        X=None, y=None, sample_weight=None, X_valid=None, y_valid=None, sample_weight_valid=None,
        cv_splits_indices=None, user_script=None,
        training_data=None, validation_data=None, label_column_name=None, weight_column_name=None,
        cv_split_column_names=None, automl_settings=None, verifier=None):
    """
    Create a dictionary with training and validation data from all supported input formats.

    :param X: Training features.
    :type X: pandas.DataFrame or numpy.ndarray or azureml.dataprep.Dataflow
    :param y: Training labels.
    :type y: pandas.DataFrame or numpy.ndarray or azureml.dataprep.Dataflow
    :param sample_weight: Sample weights for training data.
    :type sample_weight: pandas.DataFrame or numpy.ndarray or azureml.dataprep.Dataflow
    :param X_valid: validation features.
    :type X_valid: pandas.DataFrame or numpy.ndarray or azureml.dataprep.Dataflow
    :param y_valid: validation labels.
    :type y_valid: pandas.DataFrame or numpy.ndarray or azureml.dataprep.Dataflow
    :param sample_weight_valid: validation set sample weights.
    :type sample_weight_valid: pandas.DataFrame or numpy.ndarray or azureml.dataprep.Dataflow
    :param cv_splits_indices:
        Indices where to split training data for cross validation.
        Each row is a separate cross fold and within each crossfold, provide 2 arrays,
        the first with the indices for samples to use for training data and the second
        with the indices to use for validation data. i.e [[t1, v1], [t2, v2], ...]
        where t1 is the training indices for the first cross fold and v1 is the validation
        indices for the first cross fold.
    :type cv_splits_indices: numpy.ndarray
    :param training_data: The training data to be used within the experiment.
    :type training_data: typing.Union[pandas.DataFrame, azureml.core.Dataset,
        azureml.data.dataset_definition.DatasetDefinition, azureml.data.TabularDataset]
    :param validation_data: The validation data to be used within the experiment.
    :type validation_data: typing.Union[pandas.DataFrame, azureml.core.Dataset,
        azureml.data.dataset_definition.DatasetDefinition, azureml.data.TabularDataset]
    :param label_column_name: The name of the label column.
    :type label_column_name: str
    :param weight_column_name: The name of the sample weight column.
    :type weight_column_name: str
    :param cv_split_column_names: List of names for columns that contain custom cross validation split.
    :type cv_split_column_names: list(str)
    :param user_script: File path to script containing get_data()
    :param automl_settings: automl settings
    :param verifier: Verifier Manager instance.
    :type verifier: azureml.automl.runtime.faults_verifier.VerifierManager
    :return:
    """
    x_raw_column_names = None
    reset_index_msg = ("Reset dataframe index to avoid potential conflicts "
                       "if time/time_series_id_column_names column is part of an index.")

    if X is None and y is None and training_data is None:
        data_dict = _extract_user_data(user_script)
        X = data_dict.get('X')
        y = data_dict.get('y')
        sample_weight = data_dict.get('sample_weight')
        X_valid = data_dict.get('X_valid')
        y_valid = data_dict.get('y_valid')
        sample_weight_valid = data_dict.get('sample_weight_valid')
        cv_splits_indices = data_dict.get("cv_splits_indices")
        x_raw_column_names = data_dict.get("x_raw_column_names")
    elif training_data is not None and label_column_name is not None:
        if isinstance(training_data, pd.DataFrame):
            X, y, sample_weight, cv_splits_indices = _extract_data_from_combined_dataframe(
                training_data=training_data, label_column_name=label_column_name,
                sample_weight_column_name=weight_column_name, cv_split_column_names=cv_split_column_names
            )

            if validation_data is not None:
                X_valid, y_valid, sample_weight_valid, _ = _extract_data_from_combined_dataframe(
                    training_data=validation_data, label_column_name=label_column_name,
                    sample_weight_column_name=weight_column_name
                )
        elif isinstance(training_data, dprep.Dataflow):
            X, y, sample_weight, cv_splits_indices = _extract_data_from_combined_dataflow(
                training_data=training_data, label_column_name=label_column_name,
                sample_weight_column_name=weight_column_name, cv_split_column_names=cv_split_column_names
            )

            if validation_data is not None:
                X_valid, y_valid, sample_weight_valid, _ = _extract_data_from_combined_dataflow(
                    training_data=validation_data, label_column_name=label_column_name,
                    sample_weight_column_name=weight_column_name
                )
            x_raw_column_names = X.head(1).columns.values

    X = dataprep_utilities.materialize_dataflow(X)
    if X_valid is not None:
        X_valid = dataprep_utilities.materialize_dataflow(X_valid)

    y = dataprep_utilities.materialize_dataflow(y, as_numpy=True)
    if y_valid is not None:
        y_valid = dataprep_utilities.materialize_dataflow(y_valid, as_numpy=True)

    if sample_weight is not None:
        sample_weight = dataprep_utilities.materialize_dataflow(sample_weight, as_numpy=True)
    if sample_weight_valid is not None:
        sample_weight_valid = dataprep_utilities.materialize_dataflow(sample_weight_valid, as_numpy=True)

    if cv_splits_indices is not None and cv_split_column_names is None:
        # cv_splits_indices NOT derived from cv_split_column_names so it still needs to resolve
        cv_splits_indices = dataprep_utilities.resolve_cv_splits_indices(cv_splits_indices)

    # Get the raw column names
    if isinstance(X, pd.DataFrame):
        # reset index in case a customer's df contains index column(s)
        X.reset_index(inplace=True, drop=True)
        forecasting_heuristic_utils._log_warn_maybe(reset_index_msg)
        # Cache the raw column names if available
        x_raw_column_names = X.columns.values

        # reset index in case a customer's df contains index column(s)
        if X_valid is not None:
            X_valid.reset_index(inplace=True, drop=True)
            forecasting_heuristic_utils._log_warn_maybe(reset_index_msg)
    else:
        X = _data_transformation_utilities._add_raw_column_names_to_X(X)
        if X_valid is not None:
            X_valid = _data_transformation_utilities._add_raw_column_names_to_X(X_valid)

    y = _convert_to_numpy_maybe(y, 'y')
    y_valid = _convert_to_numpy_maybe(y_valid, 'y_valid')
    if isinstance(sample_weight, pd.DataFrame):
        sample_weight = sample_weight.values
    if isinstance(sample_weight_valid, pd.DataFrame):
        sample_weight_valid = sample_weight_valid.values
    if automl_settings is not None:
        # todo: This should be part of data preparation, which in turn should happen after data validations.
        #       However, currently validations happen after this step. Once we're able to flip the steps, remove
        #       the reference to common_data_validations.check_dimensions, and in general assume the underlying
        #       structures are valid (e.g. X,y have the same number of rows, have the right dimensions etc.)
        # auto cv needs the input X and y have same rows
        common_data_validations.check_dimensions(
            X=X, y=y, X_valid=X_valid, y_valid=y_valid,
            sample_weight=sample_weight, sample_weight_valid=sample_weight_valid
        )

        # When data were read try to fix the frequency.
        if automl_settings.is_timeseries and X_valid is None:
            fixed_data_set = fix_data_set_regularity_may_be(
                X, y,
                automl_settings.time_column_name,
                automl_settings.grain_column_names,
                frequency_fixer.str_to_offset_safe(automl_settings.freq,
                                                   ReferenceCodes._TRAINING_UTILITIES_CHECK_FREQ_FIX))
            X = fixed_data_set.data_x
            y = fixed_data_set.data_y
            failed = fixed_data_set.is_failed
            was_corrected = fixed_data_set.is_modified
            freq = fixed_data_set.freq
            if verifier:
                verifier.update_data_verifier_frequency_inference(failed, was_corrected)
            # Pad the short grains if needed (the short_series_handing_config_value is
            # checked by pad_short_grains_or_raise).
            X, y = short_grain_padding.pad_short_grains_or_raise(
                X, y, freq, automl_settings, ReferenceCodes._TS_ONE_VALUE_PER_GRAIN_TSUTIL, verifier)

        X, y, sample_weight, X_valid, y_valid, sample_weight_valid = rule_based_validation(
            automl_settings=automl_settings,
            X=X,
            y=y,
            sample_weight=sample_weight,
            X_valid=X_valid,
            y_valid=y_valid,
            sample_weight_valid=sample_weight_valid,
            cv_splits_indices=cv_splits_indices,
            verifier=verifier
        )

    data_dict = {
        'X': X,
        'y': y,
        'X_valid': X_valid,
        'y_valid': y_valid,
        'cv_splits_indices': cv_splits_indices,
        'x_raw_column_names': x_raw_column_names,
        'sample_weight': sample_weight,
        'sample_weight_valid': sample_weight_valid}
    return data_dict


def _convert_to_numpy_maybe(
        y: Optional[Union[np.ndarray, pd.DataFrame, pd.Series]],
        ds_name: str) -> Optional[np.ndarray]:
    """
    Try to convert y to numpy array.

    If y can not be converted to np.ndarray or has wrong shape the DataException is raised.
    :param y: The data set to be converted.
    :param ds_name: The name of the data set to convert.
    :raises: DataException
    """
    if y is None:
        return y
    if isinstance(y, pd.DataFrame):
        _check_y_shape(y, 'y')
        return cast(np.ndarray, y[y.columns[0]].values)
    if isinstance(y, pd.Series):
        return cast(np.ndarray, y.values)
    return y


def _check_y_shape(y: pd.DataFrame, ds_name: str) -> None:
    """
    Check if y data frame has only one column.

    :param y: The y dataframe.
    :param name: The name of a data set.
    :raises: DataException
    """
    if y.shape[1] > 1:
        raise DataException._with_error(AzureMLError.create(
            DataShapeMismatch, target=ds_name, reference_code=ReferenceCodes._Y_SHAPE_MISMATCH)
        )


def _get_subsampled_data_for_streaming(
        training_data: dprep.Dataflow,
        automl_settings: AutoMLBaseSettings,
        validation_data: Optional[dprep.Dataflow] = None
) -> Tuple[pd.DataFrame, np.ndarray, Optional[pd.DataFrame],
           Optional[np.ndarray], Optional[np.ndarray], Optional[np.ndarray]]:
    """
    If streaming is enabled, we do a best effort based validation of the inputs (due to potentially large data
    sizes), in order to not run out of memory.

    Sub-sampling the input Dataflows to numpy arrays this allows the validation flow (which
    can handle numpy arrays but not Dataflows directly at the moment) to proceed.

    :param training_data: The input Dataset containing training columns + target column
    :param automl_settings: AutoMLBaseSettings for this dataset
    :param validation_data: An optional validation dataset with the same schema as training_data
    :return: A materialized tuple of:
        X (pd.DataFrame), y (np.ndarray), weights (np.ndarray),
        X_valid (pd.DataFrame), y_valid (np.ndarray), weights (np.ndarray)
    """
    Validation.validate_value(
        training_data, "training_data", reference_code=ReferenceCodes._STREAMING_MISSING_TRAINING_DATA
    )
    Validation.validate_type(training_data, "training_data", expected_types=dprep.Dataflow)
    Validation.validate_value(automl_settings.label_column_name, "label_column_name")

    target_column_name = cast(str, automl_settings.label_column_name)
    validation_size = automl_settings.validation_size or 0.0
    weight_column_name = automl_settings.weight_column_name

    try:
        # validate that the Datatypes are the same for X and y (for training, validation and sample weight data)
        if validation_data is not None:
            Validation.validate_type(validation_data, "validation_data", expected_types=dprep.Dataflow)
            _validate_train_valid_datatypes(training_data, validation_data)

        X, y, sample_weight = _materialize_data_for_validation(
            training_data, target_column_name, weight_column_name)

        X_valid = None  # type: Optional[pd.DataFrame]
        y_valid = None  # type: Optional[np.ndarray]
        sample_weight_valid = None  # type: Optional[np.ndarray]

        # Don't set X_valid / y_valid in case the user specified validation_size, since in this case the
        # validation data is just a part of the overall training data.
        if validation_size == 0 and validation_data is not None:
            X_valid, y_valid, sample_weight_valid = _materialize_data_for_validation(
                training_data, target_column_name, weight_column_name)
        return X, y, X_valid, y_valid, sample_weight, sample_weight_valid
    except DprepException as e:
        logging_utilities.log_traceback(e, logger)
        dataprep_utilities.dataprep_error_handler(e)


def _materialize_data_for_validation(
        dataset: dprep.Dataflow, target_column_name: str, weight_column_name: Optional[str],
        count: int = LargeDatasetLimit.VALIDATION_SUBSAMPLE_SIZE
) -> Tuple[pd.DataFrame, np.ndarray, Optional[np.ndarray]]:
    """
    Pulls and loads a fixed number of records from the Dataflow into an in-memory pandas DataFrame, and separates
    the training and target (or weights) features (i.e. X, y, weights)

    For cases when data is remote and potentially large (e.g., streaming scenario), this method will only load a fixed
    (usually small) number of records and materialize them into an appropriate format for each of the components of
    the data. X is returned as a pandas DataFrame, while y and weights are returned as a numpy array.

    :param dataset: Pointer to the data
    :param target_column_name: The name of the column to predict
    :param weight_column_name: The name of the column that represents the weights to be assigned to each sample
    :param count: The number of records to fetch and load in-memory
    :return: A tuple of
        X (training data, pd.DataFrame), y (target column, np.ndarray), weights (weights column, np.ndarray)
    """
    Contract.assert_type(dataset, "dataset", expected_types=dprep.Dataflow)
    Contract.assert_value(target_column_name, "target_column_name")

    non_training_columns = [target_column_name]
    if weight_column_name:
        non_training_columns.append(weight_column_name)

    dataflow_X = dataset.drop_columns(non_training_columns)
    dataflow_y = dataset.keep_columns(target_column_name)
    dataflow_weights = dataset.keep_columns(weight_column_name) if weight_column_name else None

    # todo move to validators, perhaps tabular data validator constructor?
    _validate_indistinct_target_column(dataflow_X, dataflow_y)

    materialized_X = dataflow_X.take(LargeDatasetLimit.VALIDATION_SUBSAMPLE_SIZE).to_pandas_dataframe()

    Contract.assert_true(
        len(dataflow_y.head(1).columns) == 1,
        "Expected target column to be a single column, got: {}".format(len(dataflow_y.head(1).columns))
    )
    materialized_y = dataflow_y.take(count).to_pandas_dataframe().iloc[:, 0].values

    materialized_weights = None  # type: Optional[np.ndarray]
    if dataflow_weights:
        Contract.assert_true(
            len(dataflow_weights.head(1).columns) == 1,
            "Expected weights column to be a single column, got: {}".format(len(dataflow_weights.head(1).columns))
        )
        materialized_weights = dataflow_weights.take(count).to_pandas_dataframe().iloc[:, 0].values

    return materialized_X, materialized_y, materialized_weights


def _validate_indistinct_target_column(X: dprep.Dataflow, y: dprep.Dataflow) -> None:
    # todo move to validators
    # validate that column names are unique between X and y
    # (this is required, b/c we append the columns of X and y together to get a single merged Dataflow
    # that nimbus learners require as input. if X and y have shared column names, this merge throws an error)
    X_column_names = X.head(1).columns.tolist()
    y_column_name = y.head(1).columns[0]
    if y_column_name in X_column_names:
        raise DataException._with_error(AzureMLError.create(
            IndistinctLabelColumn, target='label_column_name', label_column_name=y_column_name,
            reference_code=ReferenceCodes._INDISTINCT_LABEL_COLUMN_IN_DATAFLOW)
        )


def _validate_train_valid_datatypes(training_data: dprep.Dataflow, validation_data: dprep.Dataflow) -> None:
    """
    Verifies that data types for columns in the training and validation datasets are the same

    :param training_data: The input Dataset containing training columns + target column
    :param validation_data: An optional validation dataset with the same schema as training_data
    :return None
    """
    for obj, obj_name in [(training_data, "training_data"), (validation_data, "validation_data")]:
        Contract.assert_type(obj, obj_name, expected_types=dprep.Dataflow,
                             reference_code=ReferenceCodes._STREAMING_INVALID_DATATYPE)

    if training_data is not None and validation_data is not None:
        train_dtypes = (training_data.take(LargeDatasetLimit.VALIDATION_SUBSAMPLE_SIZE).dtypes.items())
        valid_dtypes = (validation_data.take(LargeDatasetLimit.VALIDATION_SUBSAMPLE_SIZE).dtypes.items())

        for (train_col, train_datatype), (valid_col, valid_datatype) in zip(train_dtypes, valid_dtypes):
            if train_datatype != valid_datatype and train_col == valid_col:
                raise ValidationException._with_error(AzureMLError.create(
                    InconsistentColumnTypeInTrainValid, target=train_col, column_name=train_col,
                    train_dtype=train_datatype, validation_dtype=valid_datatype)
                )


def _get_raw_experiment_data(data_dict: Dict[str, Any], automl_settings: AutoMLBaseSettings) -> RawExperimentData:
    """
    Create a raw experiment data

    # 'data_dict' here contains data split into train / valid split only if:
    #   - The user provided a validation dataset explicitly
    #   - An AutoML Rule based validation was applied on the training dataset
    # For all other cases, the data is split later during consumption time (e.g., during featurization)

    :param data_dict:
    :param automl_settings:
    :return:
    """
    # For streaming cases, X & y are still populated in data_dict, only that it is the complete Dataflow
    # 'X' can also be a numpy array or a sparse matrix, if the user provided old style inputs (i.e. X and y separately)
    X = data_dict.get("X")  # type: DataInputType
    y = data_dict.get("y")  # type: Union[np.ndarray, dprep.Dataflow]
    weights = data_dict.get("sample_weight")  # type: Optional[np.ndarray]

    Validation.validate_value(X, "X")
    Validation.validate_value(y, "y")

    X_valid = data_dict.get("X_valid")  # type: Optional[DataInputType]
    y_valid = data_dict.get("y_valid")  # type: Optional[Union[np.ndarray, dprep.Dataflow]]
    weights_valid = data_dict.get("sample_weight_valid")  # type: Optional[np.ndarray]

    # The following two properties are currently only populated at this stage for streaming scenarios.
    # Non-streaming scenarios don't populate these.
    training_data = data_dict.get("training_data")  # type: Optional[dprep.Dataflow]
    validation_data = data_dict.get("validation_data")  # type: Optional[dprep.Dataflow]

    if automl_settings.enable_streaming:
        # overwrite X/y/.. with a sub-sampled and materialized version of the dataset
        X, y, X_valid, y_valid, weights, weights_valid = _get_subsampled_data_for_streaming(
            training_data, automl_settings, validation_data)

    target_column_name = automl_settings.label_column_name
    weight_column_name = automl_settings.weight_column_name
    validation_size = automl_settings.validation_size
    n_cross_validations = automl_settings.n_cross_validations
    feature_names = data_dict.get('x_raw_column_names')
    cv_splits_indices = data_dict.get('cv_splits_indices')

    X = _data_transformation_utilities._add_raw_column_names_to_X(X, feature_names)
    if X_valid is not None:
        X_valid = _data_transformation_utilities._add_raw_column_names_to_X(X_valid, feature_names)

    raw_experiment_data = RawExperimentData(X=X,
                                            y=y,
                                            target_column_name=target_column_name,
                                            weights=weights,
                                            training_data=training_data,
                                            validation_data=validation_data,
                                            X_valid=X_valid,
                                            y_valid=y_valid,
                                            weights_valid=weights_valid,
                                            feature_column_names=feature_names,
                                            weight_column_name=weight_column_name,
                                            cv_splits_indices=cv_splits_indices,
                                            validation_size=validation_size,
                                            n_cross_validations=n_cross_validations,
                                            )

    return raw_experiment_data


def validate_training_data_dict(data_dict: Dict[str, Any],
                                automl_settings: AutoMLBaseSettings,
                                check_sparse: bool = False) -> None:
    """
    Validate that training data and parameters have been correctly provided.

    :param data_dict: Key-Value pair containing information about user provided data
    :param automl_settings: AutoML configuration
    :param check_sparse:
    :return:
    """
    logger.info("Start data validation.")
    with tracer.start_as_current_span(
            TelemetryConstants.SPAN_FORMATTING.format(
                TelemetryConstants.COMPONENT_NAME, TelemetryConstants.DATA_VALIDATION
            ),
            user_facing_name=TelemetryConstants.DATA_VALIDATION_USER_FACING
    ):
        raw_experiment_data = _get_raw_experiment_data(data_dict, automl_settings)
        ml_engine.validate(raw_experiment_data, automl_settings, check_sparse)


def _construct_coord_indices_str(data_indices: np.ndarray, n_top_indices: int = 20) -> str:
    """Contruct a string with top 20 indices."""
    if data_indices.ndim == 1 or data_indices.shape[1] == 1:
        indices = sorted(data_indices)
    else:
        indices = sorted(data_indices, key=lambda x: (x[1], x[0]))
    if len(indices) <= n_top_indices:
        print_indices = data_indices  # type: Union[np.ndarray, List[List[int]]]
        return ", ".join([str(idx) for idx in print_indices])
    else:
        if data_indices.ndim == 1:
            print_indices = data_indices[:n_top_indices]
        else:
            col_idx_dict = {}  # type: Dict[int, List[List[int]]]
            for idx in indices:
                if idx[1] not in col_idx_dict:
                    col_idx_dict[idx[1]] = [idx]
                else:
                    col_idx_dict[idx[1]].append(idx)
            top_indices = sorted(col_idx_dict.keys(), key=lambda x: len(col_idx_dict[x]))
            if len(top_indices) > n_top_indices:
                print_indices = [col_idx_dict[idx][0] for idx in top_indices[:n_top_indices]]
            else:
                print_indices = [col_idx_dict[idx][0] for idx in top_indices]
        return ", ".join([str(idx) for idx in print_indices]) + "..."


def _get_data_indices_by_mask_function(data: CoreDataInputType,
                                       mask_function: 'Callable[..., Optional[Any]]') -> np.ndarray:
    """Obtain the indices list where the data entry in data has the mask function evaluated as True."""
    if isinstance(data, np.ndarray) or isinstance(data, pd.DataFrame) or isinstance(data, pd.Series):
        return np.argwhere(mask_function(data))
    elif sparse.issparse(data):
        coo_data = sparse.coo_matrix(data)
        return np.array(
            [(coo_data.row[i], coo_data.col[i]) for i in np.argwhere(mask_function(coo_data.data)).ravel()])
    else:
        return np.array([])


def _is_sparse_matrix_int_type(sparse_matrix: DataInputType) -> bool:
    """
    Check if a sparse matrix is in integer format.

    :param sparse_matrix:
    :return:
    """
    if sparse_matrix is not None and sparse.issparse(sparse_matrix):
        numpy_int_types = [np.int32, np.int64, np.int16, np.int8,
                           np.uint32, np.uint64, np.uint16, np.uint8]

        if sparse_matrix.dtype in numpy_int_types:
            return True

    return False


def _upgrade_sparse_matrix_type(sparse_matrix: DataInputType) -> DataInputType:
    """
    Convert sparse matrix in integer format into floating point format.

    This function will create a copy of the sparse matrix in when the conversion happens.
    :param sparse_matrix:
    :return:
    """
    if sparse_matrix is not None and sparse.issparse(sparse_matrix):
        if sparse_matrix.dtype == np.int32 or sparse_matrix.dtype == np.int16 or \
                sparse_matrix.dtype == np.int8 or sparse_matrix.dtype == np.uint32 or \
                sparse_matrix.dtype == np.uint16 or sparse_matrix.dtype == np.uint8 or \
                sparse_matrix.dtype == np.bool:
            return sparse_matrix.astype(np.float32)
        elif sparse_matrix.dtype == np.int64 or sparse_matrix.dtype == np.uint64:
            return sparse_matrix.astype(np.float64)
        else:
            return sparse_matrix

    return sparse_matrix


def init_client_dataset_from_fit_iteration_params(fit_iteration_parameters_dict: Dict[str, Any],
                                                  automl_settings: AutoMLBaseSettings,
                                                  cache_store: Optional[CacheStore] = None,
                                                  remote: bool = False,
                                                  init_all_stats: bool = False,
                                                  keep_in_memory: bool = False) -> ClientDatasets:
    """
    Get a ClientDatasets object from fit_iteration_parameters

    TODO: This method needs to be deprecated. ClientDatasets should be consolidated to only use transformed data ctx

    :param fit_iteration_parameters_dict: Dictionary that contains input data
    :param automl_settings:  AutoML settings config
    :param cache_store: Underlying cache store to use, will default to local FileStore
    :param remote: remote or local run flag
    :param init_all_stats: Initialize all the stats
    :param keep_in_memory: Whether to flush the data to the cache store or keep it in-memory
    :return: ClientDatasets
    """
    cv_splits = _CVSplits(X=fit_iteration_parameters_dict.get('X'),
                          y=fit_iteration_parameters_dict.get('y'),
                          frac_valid=automl_settings.validation_size,
                          cv_splits_indices=fit_iteration_parameters_dict.get('cv_splits_indices'),
                          is_time_series=automl_settings.is_timeseries,
                          timeseries_param_dict=utilities._get_ts_params_dict(automl_settings),
                          task=automl_settings.task_type)

    dataset = _get_client_dataset(fit_iteration_parameters_dict.get('X'),
                                  fit_iteration_parameters_dict.get('y'),
                                  cache_store=cache_store,
                                  sample_weight=fit_iteration_parameters_dict.get('sample_weight'),
                                  X_valid=fit_iteration_parameters_dict.get('X_valid'),
                                  y_valid=fit_iteration_parameters_dict.get('y_valid'),
                                  sample_weight_valid=fit_iteration_parameters_dict.get('sample_weight_valid'),
                                  cv_splits=cv_splits,
                                  num_classes=automl_settings.num_classes,
                                  task_type=automl_settings.task_type,
                                  y_min=automl_settings.y_min,
                                  y_max=automl_settings.y_max,
                                  init_all_stats=init_all_stats,
                                  remote=remote)

    dataset.x_raw_column_names = fit_iteration_parameters_dict.get('x_raw_column_names')

    if not keep_in_memory:
        dataset.cache_dataset(keep_in_memory)

    return dataset


def init_dataset(
        transformed_data_context: Union[TransformedDataContext, StreamingTransformedDataContext],
        cache_store: CacheStore,
        automl_settings: AutoMLBaseSettings,
        remote: bool = False,
        init_all_stats: bool = False,
        keep_in_memory: bool = False
) -> DatasetBase:
    """
    Initialize the dataset.

    :param transformed_data_context: Featurized dataset.
    :param cache_store: cache store
    :param automl_settings: automl settings
    :param remote: remote or local run flag
    :param init_all_stats: init all stats
    :param keep_in_memory: Whether to flush the data to the cache store or keep it in-memory
    :return: DatasetBase
    """
    if isinstance(transformed_data_context, StreamingTransformedDataContext):
        return init_streaming_dataset(
            transformed_data_context=transformed_data_context,
            automl_settings=automl_settings
        )

    elif isinstance(transformed_data_context, TransformedDataContext):
        return init_client_dataset(
            transformed_data_context=transformed_data_context,
            cache_store=cache_store,
            automl_settings=automl_settings,
            remote=remote,
            init_all_stats=init_all_stats,
            keep_in_memory=keep_in_memory)


def init_client_dataset(transformed_data_context: TransformedDataContext,
                        cache_store: CacheStore,
                        automl_settings: AutoMLBaseSettings,
                        remote: bool = False,
                        init_all_stats: bool = False,
                        keep_in_memory: bool = False) -> ClientDatasets:
    """
    Get the client dataset.

    :param transformed_data_context: Transformed_data_context contains X,y & other data's.
    :param cache_store: Cache store object.
    :param automl_settings: AutoML settings.
    :param remote: Remote or local run flag.
    :param init_all_stats: Initialize all the statistics.
    :param keep_in_memory: Whether to flush the data to the cache store or keep it in-memory.
    :return: ClientDatasets object.
    """
    dataset = _get_client_dataset(transformed_data_context.X,
                                  transformed_data_context.y,
                                  cache_store=cache_store,
                                  sample_weight=transformed_data_context.sample_weight,
                                  X_valid=transformed_data_context.X_valid,
                                  y_valid=transformed_data_context.y_valid,
                                  sample_weight_valid=transformed_data_context.sample_weight_valid,
                                  cv_splits=transformed_data_context.cv_splits,
                                  num_classes=automl_settings.num_classes,
                                  task_type=automl_settings.task_type,
                                  y_min=automl_settings.y_min,
                                  y_max=automl_settings.y_max,
                                  init_all_stats=init_all_stats,
                                  remote=remote,
                                  transformers=transformed_data_context.transformers,
                                  X_raw=transformed_data_context.X_raw_cleaned,
                                  y_raw=transformed_data_context.y_raw_cleaned,
                                  X_valid_raw=transformed_data_context.X_valid_raw_cleaned,
                                  y_valid_raw=transformed_data_context.y_valid_raw_cleaned)

    dataset.timeseries = transformed_data_context.timeseries
    dataset.timeseries_param_dict = transformed_data_context.timeseries_param_dict
    dataset.x_raw_column_names = transformed_data_context.x_raw_column_names
    dataset.raw_data_type = transformed_data_context._get_raw_data_type()
    dataset.raw_data_snapshot_str = transformed_data_context._get_raw_data_snapshot_str()

    if automl_settings.n_cross_validations is None and transformed_data_context.X_valid is None:
        # set the value for num_auto_cv_splits if no other mode of Validation is specified
        n_cv = transformed_data_context._get_num_cv_splits()
        dataset.num_auto_cv_splits = None if n_cv == 0 else n_cv

    if not keep_in_memory:
        dataset.cache_dataset(keep_in_memory)

    return dataset


def _get_client_dataset(X: DataInputType,
                        y: DataSingleColumnInputType,
                        cache_store: Optional[CacheStore] = None,
                        sample_weight: Optional[DataInputType] = None,
                        X_valid: Optional[DataInputType] = None,
                        y_valid: Optional[DataSingleColumnInputType] = None,
                        sample_weight_valid: Optional[DataInputType] = None,
                        cv_splits: Optional[_CVSplits] = None,
                        num_classes: Optional[int] = None,
                        task_type: str = constants.Tasks.CLASSIFICATION,
                        y_min: Optional[float] = None,
                        y_max: Optional[float] = None,
                        init_all_stats: bool = False,
                        remote: bool = True,
                        transformers: Optional[Dict[str, TransformerMixin]] = None,
                        X_raw: Optional[DataInputType] = None,
                        y_raw: Optional[DataSingleColumnInputType] = None,
                        X_valid_raw: Optional[DataInputType] = None,
                        y_valid_raw: Optional[DataSingleColumnInputType] = None) -> ClientDatasets:
    default_dataset_name = 'NoName'

    if cv_splits:
        frac_valid = cv_splits.get_fraction_validation_size()
        cv_splits_indices = cv_splits.get_custom_split_indices()
        num_cv_folds = cv_splits.get_num_k_folds()
    else:
        frac_valid = None
        cv_splits_indices = None
        num_cv_folds = None

    subsample_cache_strategy = SubsampleCacheStrategy.Classic if remote else SubsampleCacheStrategy.Preshuffle
    dataset = ClientDatasets(subsample_cache_strategy=subsample_cache_strategy, cache_store=cache_store)
    logger.info('Created ClientDataset.')

    if X_valid is not None:
        training_type = _get_training_type(
            constants.TrainingType.TrainAndValidation)

        if not (num_cv_folds == 0 or num_cv_folds is None):
            raise ConfigException._with_error(
                AzureMLError.create(
                    ConflictingValueForArguments, target="n_cross_validations",
                    arguments=', '.join(['validation_data/X_valid', 'n_cross_validations'])
                )
            )

        if not (frac_valid == 0.0 or frac_valid is None):
            raise ConfigException._with_error(
                AzureMLError.create(
                    ConflictingValueForArguments, target="validation_size",
                    arguments=', '.join(['validation_data/X_valid', 'validation_size'])
                )
            )

        if y_valid is None:
            raise ConfigException._with_error(
                AzureMLError.create(
                    ArgumentBlankOrEmpty, target="y_valid", argument_name="y_valid"
                )
            )

        logger.info('Parsing simple train validate dataset.')
        dataset.parse_simple_train_validate(name=default_dataset_name,
                                            X=X,
                                            y=y,
                                            sample_weight=sample_weight,
                                            X_valid=X_valid,
                                            y_valid=y_valid,
                                            sample_weight_valid=sample_weight_valid,
                                            task=task_type,
                                            y_min=y_min,
                                            y_max=y_max,
                                            init_all_stats=init_all_stats,
                                            transformers=transformers,
                                            X_raw=X_raw, y_raw=y_raw,
                                            X_valid_raw=X_valid_raw, y_valid_raw=y_valid_raw)

    else:
        if (num_cv_folds == 0 or num_cv_folds is None) and cv_splits_indices is None:
            training_type = _get_training_type(
                constants.TrainingType.TrainAndValidation)
        else:
            if cv_splits_indices is not None:
                num_cv_folds = len(cv_splits_indices)
            training_type = _get_training_type(
                constants.TrainingType.MeanCrossValidation, num_cv_folds)

        logger.info('Parsing cross validation dataset.')
        dataset.parse_data(name=default_dataset_name,
                           X=X,
                           y=y,
                           sample_weight=sample_weight,
                           cv_splits=cv_splits,
                           num_classes=num_classes,
                           task=task_type,
                           y_min=y_min,
                           y_max=y_max,
                           init_all_stats=init_all_stats,
                           transformers=transformers,
                           X_raw=X_raw, y_raw=y_raw,
                           X_valid_raw=X_valid_raw, y_valid_raw=y_valid_raw)

    dataset.training_type = training_type
    return dataset


def init_streaming_dataset(
        transformed_data_context: StreamingTransformedDataContext,
        automl_settings: AutoMLBaseSettings
) -> StreamingDataset:
    """
    Initialize a streaming dataset (a dataset where where all data may not fit into memory at once).

    :param transformed_data_context: The transformed data context.
    :param automl_settings: AutoML settings
    :return: A StreamingDataset
    """
    if automl_settings.label_column_name is None:
        raise ConfigException._with_error(
            AzureMLError.create(
                ArgumentBlankOrEmpty, target="label_column_name", argument_name="label_column_name"
            )
        )

    featurized_column_names = transformed_data_context.get_featurized_column_names()

    dataset_metadata = {DatasetMetadataKeys.feature_column_names: featurized_column_names,
                        DatasetMetadataKeys.label_column_name: automl_settings.label_column_name,
                        DatasetMetadataKeys.weight_column_name: automl_settings.weight_column_name,
                        DatasetMetadataKeys.raw_data_snapshot: transformed_data_context.raw_data_snapshot}

    featurization_transformer = transformed_data_context.get_featurization_transformer()

    return StreamingDataset(task=automl_settings.task_type,
                            training_data=transformed_data_context.training_data,
                            dataset_metadata=dataset_metadata,
                            validation_data=transformed_data_context.validation_data,
                            y_min=automl_settings.y_min,
                            y_max=automl_settings.y_max,
                            featurization_transformer=featurization_transformer)


def _get_training_type(training_type: str, folds: int = 0) -> str:
    """
    Determine what type of training and validation to do based on user inputs.
    """
    # TODO: make this simpler
    valid_training_types = (constants.TrainingType.TrainAndValidation,
                            constants.TrainingType.MeanCrossValidation)
    if training_type not in valid_training_types:
        raise ConfigException._with_error(
            AzureMLError.create(
                InvalidArgumentWithSupportedValues, target="training_type",
                arguments="training_type ({})".format(training_type), supported_values="%s, %s" % valid_training_types
            )
        )
    is_cv = training_type == constants.TrainingType.MeanCrossValidation
    if not ((is_cv and folds) or (not is_cv and not folds)):
        # This isn't very user friendly / actionable.
        raise ConfigException._with_error(
            AzureMLError.create(
                ConflictingValueForArguments, target="cross_validation_folds",
                arguments=', '.join(['training_type ({})'.format(constants.TrainingType.MeanCrossValidation),
                                     'cross_validation_folds'])
            )
        )
    if folds < 0 or folds == 1:
        raise ConfigException._with_error(
            AzureMLError.create(
                ArgumentOutOfRange, target="cross_validation_folds",
                argument_name="cross_validation_folds", min=2, max="inf"
            )
        )
    return training_type


def _extract_user_data(
        user_script: Any
) -> Dict[str, Optional[Union[pd.DataFrame, np.ndarray, List[str], float, List[int]]]]:
    """
    Extract data from user's module containing get_data().

    This method automatically runs during an automated machine learning experiment.
    Arguments:
        user_script {module} -- Python module containing get_data() function.

    Raises:
        DataException -- Get data script was not defined and X, y inputs were not provided.
        DataException -- Could not execute get_data() from user script.
        DataException -- Could not extract data from user script.

    Returns:
        dict -- Dictionary containing
        X_train, y_train, sample_weight, X_valid, y_valid,
        sample_weight_valid, cv_splits_indices.

    """
    Validation.validate_value(user_script, "user_script")
    try:
        output = user_script.get_data()  # type: Union[Dict[str, Any], Tuple[Any, Any, Any, Any]]
    except Exception as ex:
        msg = ("Could not execute get_data() from user script."
               "Exception: {}")
        raise UserException(msg.format(ex)).with_generic_msg(msg.format(MASKED)) from None
    if isinstance(output, dict):
        return _extract_data_from_dict(output)
    elif isinstance(output, tuple):
        return _extract_data_from_tuple(output)
    else:
        raise DataException("The output of get_data() from user script is not a dict or a tuple.", has_pii=False)


def _extract_data_from_dict(
        output: Dict[str, Any]
) -> Dict[str, Optional[Union[pd.DataFrame, np.ndarray, List[str], float, List[int]]]]:
    """
    Extract user data if it is passed as a dictionary.

    Arguments:
        output {dict} -- dictionary containing user data and metadata.

    Raises:
        DataException -- Invalid data or encountered processing issues.

    Returns:
        dict -- Dictionary containing AutoML relevant data.

    """
    X = utilities.get_value_from_dict(output, ['X'], None)
    y = utilities.get_value_from_dict(output, ['y'], None)
    sample_weight = utilities.get_value_from_dict(output, ['sample_weight'], None)
    X_valid = utilities.get_value_from_dict(output, ['X_valid'], None)
    y_valid = utilities.get_value_from_dict(output, ['y_valid'], None)
    sample_weight_valid = utilities.get_value_from_dict(
        output, ['sample_weight_valid'], None)
    X_test = utilities.get_value_from_dict(output, ['X_test'], None)
    y_test = utilities.get_value_from_dict(output, ['y_test'], None)
    data = utilities.get_value_from_dict(output, ['data_train'], None)
    columns = utilities.get_value_from_dict(output, ['columns'], None)
    label = utilities.get_value_from_dict(output, ['label'], None)
    cv_splits_indices = utilities.get_value_from_dict(
        dictionary=output,
        names=["cv_splits_indices"], default_value=None)
    x_raw_column_names = None

    if data is not None:
        if label is None and X is None and y is None:
            # This is just called in the get_data code path.
            # get_data is to soon expected to be deprecated (and has non-existent usage among customers)
            raise DataException(
                'Pandas data array received without a label. Please add a ''label'' element to the '
                'get_data() output.', has_pii=False)
        if not isinstance(label, list):
            assert (isinstance(label, str) or isinstance(label, int))
            label = [label]
        y_extracted = data[label].values
        X_extracted = data[data.columns.difference(label)]
        if columns is not None:
            X_extracted = X_extracted[X_extracted.columns.intersection(columns)]

        if X is None and y is None:
            X = X_extracted
            y = y_extracted
        else:
            if np.array_equiv(X, X_extracted.values):
                raise DataException(
                    "Different values for X and data were provided. "
                    "Please return either X and y or data and label.", has_pii=False)
            if np.array_equiv(y, y_extracted.values):
                raise DataException(
                    "Different values for y and label were provided. "
                    "Please return either X and y or data and label.", has_pii=False)
    if isinstance(X, pd.DataFrame):
        x_raw_column_names = X.columns.values
    if isinstance(y, pd.DataFrame):
        y = y.values
    if isinstance(y_valid, pd.DataFrame):
        y_valid = y_valid.values
    if isinstance(y_test, pd.DataFrame):
        y_test = y_test.values

    if X is None:
        raise DataException(
            "Could not retrieve X train data from get_data() call. "
            "Please ensure you are either returning either "
            "{X_train: <numpy array>, y_train: <numpy array>"
            "or {data: <pandas dataframe>, label: <string>", has_pii=False)
    if y is None:
        raise DataException(
            "Could not retrieve y train data from get_data() call. "
            "Please ensure you are either returning either "
            "{X_train: <numpy array>, y_train: <numpy array>"
            "or {data: <pandas dataframe>, label: <string>", has_pii=False)

    if (X_valid is None) is not (y_valid is None):
        raise DataException(
            'Received only one of X_valid or y_valid.'
            'Either both or neither value should be provided.', has_pii=False)

    return {
        "X": X,
        "y": y,
        "x_raw_column_names": x_raw_column_names,
        "sample_weight": sample_weight,
        "X_valid": X_valid,
        "y_valid": y_valid,
        "sample_weight_valid": sample_weight_valid,
        "X_test": X_test,
        "y_test": y_test,
        "cv_splits_indices": cv_splits_indices,
    }


def _extract_data_from_tuple(
        output: Tuple[Union[pd.DataFrame, np.ndarray], Union[pd.DataFrame, np.ndarray],
                      Union[pd.DataFrame, np.ndarray], Union[pd.DataFrame, np.ndarray]]
) -> Dict[str, Optional[Union[pd.DataFrame, np.ndarray, List[str], float, List[int]]]]:
    """
    Extract user data if it is passed as a tuple.

    Arguments:
        output {tuple} -- tuple containing user data.

    Raises:
        DataException -- Could not extract X, y from get_data() in user script. get_data only output {0} values.

    Returns:
        tuple -- tuple containing X_train, y_train, X_test, y_test

    """
    X_valid, y_valid = None, None
    if len(output) < 2:
        msg = ("Could not extract X, y from get_data() in user "
               "script. get_data only output {0} values.")
        raise DataException(msg.format(len(output))).with_generic_msg(msg.format(MASKED)) from None
    x_raw_column_names = None
    X = output[0]
    y = output[1]
    if isinstance(X, pd.DataFrame):
        x_raw_column_names = X.columns.values
    if isinstance(y, pd.DataFrame):
        y = y.values

    if len(output) >= 4:
        X_valid = output[2]
        y_valid = output[3]
        if isinstance(y_valid, pd.DataFrame):
            y_valid = y_valid.values

    return {
        "X": X,
        "y": y,
        "sample_weight": None,
        "x_raw_column_names": x_raw_column_names,
        "X_valid": X_valid,
        "y_valid": y_valid,
        "sample_weight_valid": None,
        "X_test": None,
        "y_test": None,
        "cv_splits_indices": None,
    }


def _extract_data_from_combined_dataframe(
        training_data: pd.DataFrame,
        label_column_name: str,
        sample_weight_column_name: Optional[str] = None,
        cv_split_column_names: Optional[List[str]] = None
) -> Tuple[pd.DataFrame, np.ndarray, Any, Any]:
    """
    Extract user data from a Pandas dataframe if it contains both training features & labels.

    :param training_data: The Pandas dataframe to extract X, y, sample_valid from.
    :param label_column_name: Name of the label column used to extract y.
    :param sample_weight_column_name: Name of the sample weight column.
    :param cv_split_column_names: List of names of the cv split columns.
    :return: A Dictionary with keys being X, y, sample_weight, cv_splits_indices extracted from training_data.
    """
    col_names_to_drop = [label_column_name]
    sample_weight = None
    cv_splits_indices = None

    if sample_weight_column_name is not None:
        col_names_to_drop.append(sample_weight_column_name)
        if sample_weight_column_name not in training_data.columns:
            raise DataException._with_error(AzureMLError.create(
                MissingColumnsInData, target='sample_weight_column_name', columns=sample_weight_column_name,
                data_object_name='training_data',
                reference_code=ReferenceCodes._EXTRACT_DATA_FROM_COMBINED_DATAFRAME_SAMPLE_WEIGHT_MISSING)
            )
        sample_weight = training_data[sample_weight_column_name].values

    if cv_split_column_names is not None:
        col_names_to_drop.extend(cv_split_column_names)
        for cv_split_column_name in cv_split_column_names:
            if cv_split_column_name not in training_data.columns:
                raise DataException._with_error(AzureMLError.create(
                    MissingColumnsInData, target='cv_split_column_names', columns=cv_split_column_name,
                    data_object_name='training_data',
                    reference_code=ReferenceCodes._EXTRACT_DATA_FROM_COMBINED_DATAFRAME_CV_SPLIT_COLUMNS_MISSING)
                )
        cv_split_columns = training_data[cv_split_column_names]
        cv_splits_indices = _convert_cv_split_columns_to_cv_splits_indices(cv_split_columns)

    X = training_data[training_data.columns.difference(col_names_to_drop)]
    if label_column_name not in training_data.columns:
        raise DataException._with_error(
            AzureMLError.create(MissingColumnsInData, target='X', columns=label_column_name, data_object_name='X')
        )
    y = training_data[label_column_name].values

    return X, y, sample_weight, cv_splits_indices


def _extract_data_from_combined_dataflow(
        training_data: dprep.Dataflow,
        label_column_name: str,
        sample_weight_column_name: Optional[str] = None,
        cv_split_column_names: Optional[List[str]] = None,
        validate_columns_exist: bool = True
) -> Tuple[Any, Any, Any, Any]:
    """
    Extract user data from a Dataflow if it contains both training features & labels.

    :param training_data: The Dataflow to extract X, y, sample_valid from.
    :param label_column_name: Name of the label column used to extract y.
    :param sample_weight_column_name: Name of the sample weight column.
    :param cv_split_column_names: List of names of the cv split columns.
    :return: A Dictionary with keys being X, y, sample_weight, cv_splits_indices extracted from training_data.
    """
    col_names_to_drop = []
    if label_column_name is not None:
        col_names_to_drop.append(label_column_name)
    sample_weight = None
    cv_splits_indices = None

    if sample_weight_column_name is not None:
        col_names_to_drop.append(sample_weight_column_name)
        try:
            sample_weight = training_data.keep_columns([sample_weight_column_name],
                                                       validate_column_exists=validate_columns_exist)
        except DataflowValidationError:
            raise DataException._with_error(AzureMLError.create(
                MissingColumnsInData, target="weight_column_name", data_object_name="X",
                columns="{} (weight_column_name)".format(sample_weight_column_name),
                reference_code=ReferenceCodes._EXTRACT_DATA_FROM_COMBINED_DATAFLOW_SAMPLE_WEIGHT_MISSING)
            )

    if cv_split_column_names is not None:
        cv_split_column_names = list(filter(None, cv_split_column_names))
        col_names_to_drop.extend(cv_split_column_names)
        try:
            cv_split_columns = training_data.keep_columns(cv_split_column_names,
                                                          validate_column_exists=validate_columns_exist)
            if cv_split_columns.row_count:
                cv_splits_indices = _convert_cv_split_columns_to_cv_splits_indices(cv_split_columns)
        except DataflowValidationError:
            raise DataException._with_error(AzureMLError.create(
                MissingColumnsInData, target="cv_split_column_names", data_object_name="X",
                columns="{} (cv_split_column_names)".format(cv_split_column_names),
                reference_code=ReferenceCodes._EXTRACT_DATA_FROM_COMBINED_DATAFLOW_CV_SPLIT_COLUMNS_MISSING)
            )

    X = training_data.drop_columns(col_names_to_drop)
    try:
        y = training_data.keep_columns([label_column_name],
                                       validate_column_exists=validate_columns_exist) \
            if label_column_name is not None else None
    except DataflowValidationError:
        raise DataException._with_error(AzureMLError.create(
            MissingColumnsInData, target="label_column_name", data_object_name="X",
            columns="{} (label_column_name)".format(label_column_name),
            reference_code=ReferenceCodes._EXTRACT_DATA_FROM_COMBINED_DATAFLOW_LABEL_COLUMN_MISSING)
        )
    return X, y, sample_weight, cv_splits_indices


def _convert_cv_split_columns_to_cv_splits_indices(
        cv_split_columns: Union[dprep.Dataflow, pd.DataFrame]
) -> List[List[Any]]:
    cv_splits_indices = []
    if isinstance(cv_split_columns, pd.DataFrame):
        cv_split_columns_numpy = cv_split_columns.values
    else:
        cv_split_columns_numpy = dataprep_utilities.materialize_dataflow(cv_split_columns, as_numpy=True)
    if cv_split_columns_numpy.ndim == 1:
        training_indices, validation_indices = _get_column_to_cv_splits_indices(cv_split_columns_numpy)
        cv_splits_indices.append([training_indices, validation_indices])
    else:
        for i in range(cv_split_columns_numpy.shape[1]):
            column = cv_split_columns_numpy[:, i]
            training_indices, validation_indices = _get_column_to_cv_splits_indices(column)
            cv_splits_indices.append([training_indices, validation_indices])
    return cv_splits_indices


def _get_column_to_cv_splits_indices(column: np.ndarray) -> Tuple[Any, Any]:
    try:
        training_indices, = np.where(column.astype(int) == 1)
        validation_indices, = np.where(column.astype(int) == 0)

        # verify number of indices extracted equal number of total rows
        if len(training_indices) + len(validation_indices) != len(column):
            raise ValueError("cv split column had values outside of the set {0, 1}")
    except ValueError as ve:
        raise DataException._with_error(
            AzureMLError.create(InvalidValuesInCVSplitColumn, target="cv_split_column_names",
                                reference_code=ReferenceCodes._CV_SPLIT_COLUMN_CONVERSION_INCORRECT_INT_VALUE),
            inner_exception=ve
        ) from ve

    return training_indices, validation_indices


def _check_if_automl_model_is_explainable(automl_algo_tag: str, is_ensemble: bool = False) -> bool:
    unexplainable_models = [constants.SupportedModels.Forecasting.Average,
                            constants.SupportedModels.Forecasting.Naive,
                            constants.SupportedModels.Forecasting.SeasonalAverage,
                            constants.SupportedModels.Forecasting.SeasonalNaive,
                            constants.SupportedModels.Forecasting.TCNForecaster,
                            constants._PrivateModels.Forecasting.ExponentialSmoothing]
    explainable = True
    if is_ensemble:
        # Parse the tag assuming it is a string representation of a Python list
        ensemble_algo_list = ast.literal_eval(automl_algo_tag)
        if len(set(unexplainable_models).intersection(ensemble_algo_list)) > 0:
            explainable = False
    else:
        if automl_algo_tag in unexplainable_models:
            explainable = False

    return explainable


def _get_model_exp_property(automl_run: Any) -> bool:
    try:
        from azureml.train.automl.runtime._remote_script import model_exp_wrapper
        automl_algo_name = automl_run.properties.get('run_algorithm')
        if automl_algo_name != 'StackEnsemble' and automl_algo_name != 'VotingEnsemble':
            if_model_is_explainable = \
                _check_if_automl_model_is_explainable(automl_algo_name, is_ensemble=False)
            if logger is not None and not if_model_is_explainable:
                logger.warning(automl_algo_name + ' is not explainable for AutoML run ' + str(automl_run.id))

            return if_model_is_explainable
        else:
            ensemble_algo_names_list_str = automl_run.tags.get('ensembled_algorithms')
            if ensemble_algo_names_list_str is not None:
                if_ensemble_model_is_explainable = \
                    _check_if_automl_model_is_explainable(ensemble_algo_names_list_str, is_ensemble=True)
                if logger is not None and not if_ensemble_model_is_explainable:
                    msg = ('{} with constituent models {} is not explainable for AutoML run {}'
                           .format(automl_algo_name, ensemble_algo_names_list_str, automl_run.id))
                    logger.warning(msg)

                return if_ensemble_model_is_explainable
            else:
                return True
    except Exception:
        return False
